/* Copyright (C) 2019 Terence Kelly.  All rights reserved.
   Failure-atomic msync() in user space via snapshots:  Implementation. */

#define _POSIX_C_SOURCE 200809L

#include <stddef.h>
#include <unistd.h>
#include <linux/fs.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <sys/types.h>
#include "famus_snap.h"

static int rwperm(mode_t m, unsigned int r, unsigned int w) {
  return (!!(m & S_IRUSR) == r) && (!!(m & S_IWUSR) == w);
}

#define  L __LINE__
#define RL return L  /* indicates where error occurs */

/* We must fsync() backing file twice to ensure that snapshot
   data are durable before success indicator (file permission)
   becomes durable.  We're not using fallocate() to reserve
   space for worst-case scenario in which backing file and
   snapshot file diverge completely, because that could defeat
   the reflink sharing that makes snapshots efficient; read
   "man ioctl_ficlone".  The "ioctl(FICLONE)" works only on
   reflink-enabled file systems, e.g., BtrFS, XFS, OCFS2. */
int famus_snap_sync(fd_t bfd, fd_t snapfd, fd_t dirfd) {
  struct stat sb;
  if (0 != fstat(snapfd, &sb))           RL;
  if (! rwperm(sb.st_mode, 0, 1))        RL;
  if (0 != ioctl(snapfd, FICLONE, bfd))  RL;
  if (0 != fsync(snapfd))                RL;
  if (0 != fchmod(snapfd, S_IRUSR))      RL;
  if (0 != fstat(snapfd, &sb))           RL;  /* paranoia */
  if (! rwperm(sb.st_mode, 1, 0))        RL;  /* paranoia */
  if (0 != fsync(snapfd))                RL;
  if (0 < dirfd && 0 != fsync(dirfd))    RL;
  if (0 != close(snapfd))                RL;
  return 0;
}

#define RN return NULL

void * famus_snap_map(void * addr, size_t * plen, int flags,
                      fd_t bfd, fd_t snapfd, fd_t dirfd,
                      int * status) {
  struct stat sb;  void *a;  int prot = PROT_READ | PROT_WRITE;
  if (NULL == status)            {              RN; }
  if (NULL == plen)              { *status = L; RN; }
  if (0 == (flags & MAP_SHARED)) { *status = L; RN; }
  if (0 != fstat(bfd, &sb))      { *status = L; RN; }
  *plen = (size_t)sb.st_size;
  a = mmap(addr, *plen, prot, flags, bfd, 0);
  if (MAP_FAILED == a)           { *status = L; RN; }
  if (NULL == a) {
    if (0 != munmap(a, *plen))     *status = L;
    else                           *status = L;
    RN;
  }
  if (0 != (*status = famus_snap_sync(bfd, snapfd, dirfd))) {
    if (0 != munmap(a, *plen))     *status = L;
    RN;
  }
  return a;
}
