/* Copyright (C) 2019 Terence Kelly.  All rights reserved.
   Failure-atomic msync() in user space via snapshots:  Interface.

   Notes:

   Previous implementations of failure-atomic msync() guarantee that
   the backing file always reflects the most recent successful
   msync() call, even in the presence of failures (e.g., power
   outages, OS kernel panics, application process crashes).

   By contrast, this implementation guarantees that a caller-
   specified file will contain a snapshot reflecting the state of the
   backing file's in-memory image at the most recent successful call
   of our mmap() or msync() analogue.  Following a crash, recovery
   should discard the backing file and recover from the most recent
   readable snapshot file.

   The caller should supply a new snapshot file descriptor for the
   initial famus_snap_map() call and for each famus_snap_sync() call;
   after the call succeeds the caller may delete the previous
   snapshot file.  "Recycling" a snapshot file is not safe.  The
   caller's snapshot file descriptor should be opened for writing but
   not reading.  Upon successful return the snapshot file will be
   readable but not writable and the snapshot file descriptor will
   have been closed.

   To ensure that a file is visible after a crash, it is sometimes
   necessary to fsync() not only the file but also the directory that
   contains it and all of *its* parent directories up to the root
   directory, "/".  The mmap() and msync() analogues below will
   optionally fsync() a caller-supplied directory fd; it is the
   caller's responsibility to perform any other fsync() calls that
   may be necessary.

   The implementation relies on efficient reflink-based file
   snapshots via ioctl(FICLONE), which is supported only in a few
   file systems, e.g., BtrFS and XFS.  Reflink-based snapshots use a
   copy-on-write mechanism that may cause mysterious failures if
   storage is exhausted *outside* of calls to the functions below.
*/

#ifndef FAMUS_SNAP_H_INCLUDED
#define FAMUS_SNAP_H_INCLUDED

typedef const int fd_t;  /* file descriptor type */

/* Analogue of mmap().  Returns NULL if any failure occurs; debugging
   info returned via status argument.  Backing file is mapped with
   read/write permission.  Recommended practice is to provide fd of
   directory containing snapshot file; the directory fd will be
   fsync()'d to ensure that the snapshot file is visible to
   post-crash recovery code.  To delete mapping, use ordinary
   munmap() with address & length returned by this function. */
extern void * famus_snap_map(
  void   * addr,  /* the usual address hint passed to mmap() */
  size_t * plen,  /* set to backing file size for convenience */
  int     flags,  /* passed to mmap(); must contain MAP_SHARED */
  fd_t      bfd,  /* backing file to be mapped by mmap() */
  fd_t   snapfd,  /* file where initial snapshot will be created;
                     should be new empty file opened for writing but
                     not reading by user */
  fd_t    dirfd,  /* fd of directory containing snapshot file;
                     if >0 will be fsync()'d, else ignored */
  int  * status   /* diagnostic information returned here;
                     zero iff success */
);

/* Analogue of msync().  Returns debugging info. */
extern int famus_snap_sync(
  fd_t    bfd,  /* backing file mapped by famus_snap_map() */
  fd_t snapfd,  /* file where next snapshot will be created; same
                   rules as for famus_snap_map() */
  fd_t  dirfd   /* same role as in famus_snap_map() */
);

#endif

